// src/lib.ts
import {
  cursorCharLeft,
  cursorCharRight,
  cursorDocEnd,
  cursorDocStart,
  cursorLineBoundaryRight,
  cursorLineDown,
  cursorLineStart,
  cursorLineUp,
  cursorPageDown,
  cursorPageUp,
  deleteCharBackward,
  deleteCharForward,
  insertNewlineAndIndent,
  selectAll,
  selectCharLeft,
  selectCharRight,
  selectDocEnd,
  selectDocStart,
  selectGroupLeft,
  selectGroupRight,
  selectLineBoundaryRight,
  selectLineDown,
  selectLineStart,
  selectLineUp,
  selectPageDown,
  selectPageUp,
  selectParentSyntax,
  toggleComment
} from "@codemirror/commands";
import {
  EditorSelection as EditorSelection2,
  EditorState as EditorState2,
  StateField as StateField2,
  Transaction
} from "@codemirror/state";
import {
  Decoration,
  EditorView,
  WidgetType,
  drawSelection,
  getPanel,
  keymap,
  showPanel
} from "@codemirror/view";
import {
  SearchQuery,
  getSearchQuery,
  search,
  setSearchQuery
} from "@codemirror/search";
import { matchBrackets } from "@codemirror/language";

// src/state.ts
import {
  StateEffect,
  StateField
} from "@codemirror/state";
var modeEffect = StateEffect.define();
var modeField = StateField.define({
  create() {
    return { type: 0 /* Normal */, minor: 2 /* Normal */ };
  },
  update(mode, tr) {
    for (const effect of tr.effects) {
      if (effect.is(modeEffect)) {
        mode = effect.value;
      }
    }
    return mode;
  }
});
var yankEffect = StateEffect.define();
var registerField = StateField.define({
  create() {
    return "";
  },
  update(register, tr) {
    for (const effect of tr.effects) {
      if (effect.is(yankEffect)) {
        register = effect.value;
      }
    }
    return register;
  }
});
var searchRegisterField = StateField.define({
  create() {
    return { active: null };
  },
  update(search2, tr) {
    for (const effect of tr.effects) {
      if (effect.is(searchEffect)) {
        const effectValue = effect.value;
        switch (effectValue.type) {
          case 0 /* Start */: {
            search2 = { ...search2, original: effectValue.selection };
            break;
          }
          case 1 /* Exit */: {
            search2 = {
              original: void 0,
              active: effectValue.query ?? search2.active
            };
            break;
          }
        }
      }
    }
    return search2;
  }
});
var searchEffect = StateEffect.define();
var historyEffect = StateEffect.define();
var historyField = StateField.define({
  create() {
    return { checkpoints: [], cursor: -1, pending: null, head: null };
  },
  update(value, tr) {
    for (const effect of tr.effects) {
      if (effect.is(historyEffect)) {
        const effectValue = effect.value;
        switch (effectValue.type) {
          case "move": {
            const cursor = value.cursor + effectValue.offset;
            value = {
              checkpoints: value.checkpoints,
              pending: null,
              cursor,
              head: value.head ?? effectValue.head ?? null
            };
            break;
          }
          case "add": {
            if (false) {
              if (value.pending && effectValue.temp) {
                throw new Error("Unexpected temp");
              }
            }
            const checkpoints = value.cursor === -1 ? value.checkpoints : value.checkpoints.slice(value.cursor + 1);
            value = effectValue.temp ? {
              checkpoints: value.checkpoints,
              cursor: value.cursor,
              pending: effectValue.state,
              head: value.head
            } : {
              checkpoints: [effectValue.state, ...checkpoints],
              cursor: -1,
              pending: null,
              head: null
            };
            break;
          }
          case "commit": {
            if (false) {
              if (!value.pending) {
                throw new Error("unexpected commit");
              }
            }
            if (effectValue.state.doc.eq(value.pending.doc)) {
              value = {
                ...value,
                pending: null
              };
              break;
            }
            const checkpoints = value.cursor === -1 ? value.checkpoints : value.checkpoints.slice(value.cursor + 1);
            value = {
              checkpoints: [value.pending, ...checkpoints],
              cursor: -1,
              pending: null,
              head: null
            };
            break;
          }
        }
      }
    }
    return value;
  }
});

// src/lib.ts
var MODE_EFF = {
  NORMAL: modeEffect.of({
    type: 0 /* Normal */,
    minor: 2 /* Normal */
  }),
  NORMAL_GOTO: modeEffect.of({
    type: 0 /* Normal */,
    minor: 3 /* Goto */
  }),
  NORMAL_MATCH: modeEffect.of({
    type: 0 /* Normal */,
    minor: 5 /* Match */
  }),
  SELECT: modeEffect.of({
    type: 4 /* Select */,
    minor: 2 /* Normal */
  }),
  SELECT_GOTO: modeEffect.of({
    type: 4 /* Select */,
    minor: 3 /* Goto */
  }),
  SELECT_MATCH: modeEffect.of({
    type: 4 /* Select */,
    minor: 5 /* Match */
  }),
  INSERT: modeEffect.of({ type: 1 /* Insert */ })
};
function moveDown(view, mode) {
  mode.type === 0 /* Normal */ ? cursorLineDown(view) : selectLineDown(view);
}
function moveUp(view, mode) {
  mode.type === 0 /* Normal */ ? cursorLineUp(view) : selectLineUp(view);
}
function moveLeft(view, mode) {
  mode.type === 0 /* Normal */ ? cursorCharLeft(view) : selectCharLeft(view);
}
function moveRight(view, mode) {
  mode.type === 0 /* Normal */ ? cursorCharRight(view) : selectCharRight(view);
}
function addSearch(view, query) {
  const searchRegister = view.state.field(searchRegisterField);
  const match = query.getCursor(
    view.state,
    (searchRegister.original ?? view.state.selection).main.to
  ).next();
  if (!match.done) {
    view.dispatch({
      selection: EditorSelection2.range(match.value.from, match.value.to),
      scrollIntoView: true
    });
  } else {
    view.dispatch({
      selection: searchRegister.original,
      scrollIntoView: true
    });
  }
}
var helixCommandBindings = {
  insert: {
    Backspace(view) {
      deleteCharBackward(view);
    },
    Delete(view) {
      deleteCharForward(view);
    },
    Enter(view) {
      insertNewlineAndIndent(view);
    },
    Escape(view) {
      view.dispatch({
        effects: [
          MODE_EFF.NORMAL,
          historyEffect.of({ type: "commit", state: view.state })
        ]
      });
    }
  },
  normal: {
    // this one is special: we let it apply to all other minor modes
    Escape(view, mode_) {
      const mode = mode_;
      if (mode.type === 0 /* Normal */ && mode.minor === 2 /* Normal */ && mode.expecting == null) {
        return true;
      }
      view.dispatch({
        effects: [MODE_EFF.NORMAL]
      });
      if (mode.expecting) {
        const panel = getPanel(view, commandPanel);
        panel.showCommand(null);
      }
    },
    ["/"](view) {
      view.dispatch({
        effects: searchEffect.of({
          type: 0 /* Start */,
          selection: view.state.selection
        })
      });
      view.dispatch({
        effects: setSearchQuery.of(new SearchQuery({ search: "" }))
      });
      const panel = getPanel(view, commandPanel);
      panel?.showSearchInput();
    },
    ["y"](view) {
      const selection = view.state.selection.main;
      const range = helixSelection(selection, view.state.doc);
      view.dispatch({
        effects: yankEffect.of(view.state.doc.slice(range.from, range.to))
      });
    },
    ["a"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection2.cursor(selection.to)
        });
      }
    },
    ["A"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const end = view.state.doc.lineAt(selection.to).to;
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection2.cursor(end)
        });
      }
    },
    ["c"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const range = helixSelection(selection, view.state.doc);
        view.dispatch({
          effects: [
            yankEffect.of(view.state.doc.slice(range.from, range.to)),
            MODE_EFF.INSERT
          ],
          changes: {
            from: range.from,
            to: range.to,
            insert: ""
          }
        });
      }
    },
    ["d"]: {
      checkpoint: true,
      command(view) {
        const selection = view.state.selection.main;
        const range = helixSelection(selection, view.state.doc);
        view.dispatch({
          effects: [
            yankEffect.of(view.state.doc.slice(range.from, range.to)),
            MODE_EFF.NORMAL
          ],
          changes: {
            from: range.from,
            to: range.to
          }
        });
      }
    },
    ["P"]: {
      checkpoint: true,
      command(view) {
        const range = view.state.selection.main;
        const yanked = view.state.field(registerField);
        const spec = { from: range.from, insert: yanked };
        const change = view.state.changes(spec);
        view.dispatch(
          { changes: change },
          {
            selection: { anchor: range.from, head: range.from + yanked.length },
            sequential: true
          },
          { effects: MODE_EFF.NORMAL }
        );
      }
    },
    ["p"]: {
      checkpoint: true,
      command(view) {
        const range = view.state.selection.main;
        const yanked = view.state.field(registerField);
        const spec = { from: range.to, insert: yanked };
        const change = view.state.changes(spec);
        view.dispatch(
          { changes: change },
          {
            selection: { anchor: range.to, head: range.to + yanked.length },
            sequential: true
          },
          { effects: MODE_EFF.NORMAL }
        );
      }
    },
    ["R"]: {
      checkpoint: true,
      command(view) {
        const tr = view.state.replaceSelection(view.state.field(registerField));
        tr.effects = MODE_EFF.NORMAL;
        view.dispatch(tr);
      }
    },
    ["w"](view, mode) {
      if (mode.type === 0 /* Normal */) {
        const current = view.state.selection.main;
        view.dispatch({
          selection: EditorSelection2.single(Math.max(current.from, current.to))
        });
      }
      return selectGroupRight(view);
    },
    ["b"](view, mode) {
      if (mode.type === 0 /* Normal */) {
        const current = view.state.selection.main;
        view.dispatch({
          selection: EditorSelection2.single(Math.min(current.from, current.to))
        });
      }
      return selectGroupLeft(view);
    },
    ["v"](view, mode) {
      view.dispatch({
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.SELECT : MODE_EFF.NORMAL
      });
    },
    ["g"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL_GOTO : MODE_EFF.SELECT_GOTO
      });
    },
    ["m"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL_MATCH : MODE_EFF.SELECT_MATCH
      });
    },
    ["i"]: {
      checkpoint: "temp",
      command(view) {
        const selection = view.state.selection.main;
        const cursor = Math.min(selection.from, selection.to);
        view.dispatch({
          effects: MODE_EFF.INSERT,
          selection: EditorSelection2.cursor(cursor)
        });
      }
    },
    ["h"](view, mode) {
      return moveLeft(view, mode);
    },
    ["j"]: moveDown,
    ["ArrowDown"]: moveDown,
    ["ArrowUp"]: moveUp,
    ["k"]: moveUp,
    ["ArrowRight"]: moveRight,
    ["ArrowLeft"]: moveLeft,
    ["l"](view, mode) {
      return moveRight(view, mode);
    },
    ["%"](view) {
      return selectAll(view);
    },
    ["o"]: {
      checkpoint: "temp",
      command(view) {
        insertLineAndEdit(view, true);
      }
    },
    ["O"]: {
      checkpoint: "temp",
      command(view) {
        insertLineAndEdit(view, false);
      }
    },
    ["f"](view, mode) {
      setFindMode(view, "f", mode, {
        inclusive: true,
        reverse: false
      });
    },
    ["F"](view, mode) {
      setFindMode(view, "F", mode, {
        inclusive: true,
        reverse: true
      });
    },
    ["t"](view, mode) {
      setFindMode(view, "t", mode, {
        inclusive: false,
        reverse: false
      });
    },
    ["T"](view, mode) {
      setFindMode(view, "T", mode, {
        inclusive: false,
        reverse: true
      });
    },
    ["u"](view) {
      const { checkpoints, cursor } = view.state.field(historyField);
      const nextCursor = cursor + 1;
      const state = checkpoints[nextCursor];
      if (!state) {
        return true;
      }
      view.dispatch({
        changes: {
          from: 0,
          to: view.state.doc.length,
          insert: state.doc
        },
        selection: state.selection,
        effects: historyEffect.of({
          type: "move",
          offset: 1,
          head: view.state
        }),
        scrollIntoView: true
      });
    },
    ["U"](view) {
      const { checkpoints, cursor, head } = view.state.field(historyField);
      const nextCursor = cursor - 1;
      const state = nextCursor === -1 ? head : checkpoints[nextCursor];
      if (!state) {
        return true;
      }
      view.dispatch({
        changes: {
          from: 0,
          to: view.state.doc.length,
          insert: state.doc
        },
        selection: state.selection,
        effects: historyEffect.of({
          type: "move",
          offset: -1
        }),
        scrollIntoView: true
      });
    },
    ["x"](view) {
      const initial = view.state.selection.main;
      const anchorLine = view.state.doc.lineAt(initial.anchor);
      const headLine = view.state.doc.lineAt(initial.head);
      const [startLine, endLine] = anchorLine.number < headLine.number ? [anchorLine, headLine] : [headLine, anchorLine];
      const ideal = EditorSelection2.range(startLine.from, endLine.to);
      let nextSel;
      if (ideal.from === initial.from && ideal.to === initial.to && view.state.doc.lines > endLine.number) {
        const nextLine = view.state.doc.line(endLine.number + 1);
        nextSel = EditorSelection2.range(startLine.from, nextLine.to);
      } else {
        nextSel = EditorSelection2.range(startLine.from, endLine.to);
      }
      view.dispatch({
        selection: nextSel
      });
      return true;
    },
    ["n"](view) {
      const register = view.state.field(searchRegisterField);
      const active = register.active;
      if (!active?.valid) {
        return true;
      }
      const cursor = active.getCursor(view.state, view.state.selection.main.to);
      let match = cursor.next();
      if (match.done) {
        match = active.getCursor(view.state).next();
      }
      if (!match.done) {
        view.dispatch({
          selection: EditorSelection2.range(match.value.from, match.value.to),
          scrollIntoView: true
        });
      }
    },
    ["N"](view) {
      const register = view.state.field(searchRegisterField);
      const active = register.active;
      if (!active?.valid) {
        return true;
      }
      const cursor = active.getCursor(view.state);
      const selection = view.state.selection.main;
      let match;
      let before = false;
      for (const item of {
        [Symbol.iterator]() {
          return cursor;
        }
      }) {
        if (item.to < selection.from) {
          match = item;
          before = true;
        } else if (item.from > selection.to) {
          if (before) {
            break;
          } else {
            match = item;
          }
        }
      }
      if (match) {
        view.dispatch({
          selection: EditorSelection2.range(match.from, match.to),
          scrollIntoView: true
        });
      }
    },
    ["Ctrl-d"](view, mode) {
      mode.type === 0 /* Normal */ ? cursorPageDown(view) : selectPageDown(view);
    },
    ["PageDown"](view, mode) {
      mode.type === 0 /* Normal */ ? cursorPageDown(view) : selectPageDown(view);
    },
    ["PageUp"](view, mode) {
      mode.type === 0 /* Normal */ ? cursorPageUp(view) : selectPageUp(view);
    },
    ["Ctrl-u"](view, mode) {
      mode.type === 0 /* Normal */ ? cursorPageUp(view) : selectPageUp(view);
    },
    [";"](view) {
      const selection = view.state.selection.main;
      view.dispatch({
        selection: EditorSelection2.cursor(selection.head)
      });
    },
    ["Alt-;"](view) {
      const selection = view.state.selection.main;
      view.dispatch({
        selection: EditorSelection2.range(selection.head, selection.anchor)
      });
    },
    ["Alt-ArrowUp"](view) {
      return selectParentSyntax(view);
    },
    ["Ctrl-c"]: {
      checkpoint: true,
      command(view) {
        toggleComment(view);
        view.original.dispatch({
          effects: MODE_EFF.NORMAL
        });
      }
    }
  },
  goto: {
    ["g"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      isNormal ? cursorDocStart(view) : selectDocStart(view);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["e"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      isNormal ? cursorDocEnd(view) : selectDocEnd(view);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    ["h"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      isNormal ? cursorLineStart(view) : selectLineStart(view);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    },
    j: moveDown,
    k: moveUp,
    ["l"](view, mode) {
      const isNormal = mode.type === 0 /* Normal */;
      isNormal ? cursorLineBoundaryRight(view) : selectLineBoundaryRight(view);
      view.dispatch({
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      });
    }
  },
  match: {
    ["s"]: {
      checkpoint: true,
      command(view, mode) {
        view.dispatch({
          effects: modeEffect.of({
            type: mode.type,
            minor: 5 /* Match */,
            expecting: {
              callback: surround,
              metadata: view
            }
          })
        });
      }
    },
    ["m"](view, mode) {
      const selection = matchBracket(view) ?? void 0;
      const isNormal = mode.type === 0 /* Normal */;
      view.dispatch({
        selection: selection?.from == null ? void 0 : isNormal ? EditorSelection2.cursor(selection.from) : EditorSelection2.range(
          view.state.selection.main.anchor,
          selection.from
        ),
        effects: isNormal ? MODE_EFF.NORMAL : MODE_EFF.SELECT,
        scrollIntoView: true
      });
    }
  }
};
function setFindMode(view, status, mode, metadata) {
  const effect = modeEffect.of({
    type: mode.type,
    minor: 2 /* Normal */,
    expecting: {
      callback: findText,
      metadata
    }
  });
  const panel = getPanel(view, commandPanel);
  panel.showCommand(status);
  view.dispatch({ effects: effect });
}
function helixSelection(range, doc) {
  const to = Math.min(range.to + 1, doc.length);
  const forward = range.head >= range.anchor;
  const anchor = forward ? range.anchor : to;
  const head = forward ? to : range.head;
  return EditorSelection2.range(anchor, head);
}
function insertLineAndEdit(view, below) {
  let from;
  let cursor;
  if (below) {
    const line = view.state.doc.lineAt(view.state.selection.main.to);
    from = line.to;
    cursor = from + 1;
  } else {
    const line = view.state.doc.lineAt(view.state.selection.main.from);
    from = line.from;
    cursor = from;
  }
  view.dispatch({
    changes: {
      from,
      insert: "\n"
    },
    selection: EditorSelection2.cursor(cursor),
    effects: MODE_EFF.INSERT
  });
}
function toCodemirrorKeymap(keybindings) {
  const allKeys = [
    ...new Set(
      Object.values(keybindings).flat().flatMap((binding) => Object.keys(binding))
    )
  ];
  function apply(def, view, mode) {
    if (typeof def === "function") {
      return def(view, mode);
    } else {
      const temp = def.checkpoint === "temp";
      return def.command(
        {
          original: view,
          dispatch(...args) {
            view.dispatch(
              {
                effects: historyEffect.of({
                  type: "add",
                  state: view.state,
                  temp
                })
              },
              ...args
            );
          },
          get state() {
            return view.state;
          }
        },
        mode
      );
    }
  }
  const codemirrorKeybindings = [];
  for (const key of allKeys) {
    const insertCommand = keybindings.insert[key];
    const normalCommand = keybindings.normal[key];
    const gotoCommand = keybindings.goto[key];
    const matchCommand = keybindings.match[key];
    const esc = key === "Escape";
    const isChar = key.length === 1;
    const command = (view) => {
      const mode = view.state.field(modeField);
      if (mode.type === 1 /* Insert */) {
        if (insertCommand) {
          return insertCommand(view, void 0) ?? true;
        } else {
          return false;
        }
      }
      if (mode.expecting && isChar) {
        return false;
      }
      let result;
      if (esc || mode.minor === 2 /* Normal */ && normalCommand) {
        result = apply(normalCommand, view, mode);
      } else if (mode.minor === 3 /* Goto */ && gotoCommand) {
        result = apply(gotoCommand, view, mode);
      } else if (mode.minor === 5 /* Match */ && matchCommand) {
        result = apply(matchCommand, view, mode);
      } else {
        return false;
      }
      return result ?? true;
    };
    codemirrorKeybindings.push({
      key,
      run: command
    });
  }
  return codemirrorKeybindings;
}
var EndLineCursor = class extends WidgetType {
  toDOM() {
    const span = document.createElement("span");
    span.classList.add("cm-hx-cursor");
    span.classList.add("cm-hx-cursor-endline");
    span.textContent = " ";
    return span;
  }
};
var cursorMark = Decoration.mark({ class: "cm-hx-cursor" });
var endlineCursorWidget = Decoration.widget({ widget: new EndLineCursor() });
var cursorField = StateField2.define({
  create(state) {
    return drawCursorMark(state.selection, state.doc);
  },
  update(_value, tr) {
    return drawCursorMark(tr.newSelection, tr.newDoc);
  },
  provide(field) {
    return EditorView.decorations.from(field);
  }
});
function drawCursorMark(selection, doc) {
  const head = selection.main.head;
  const line = doc.lineAt(head);
  if (line.to === head) {
    return Decoration.set(
      endlineCursorWidget.range(selection.main.head, selection.main.head)
    );
  } else {
    return Decoration.set(
      cursorMark.range(selection.main.head, selection.main.head + 1)
    );
  }
}
function letThrough(tr) {
  return tr;
}
var changeFilter = EditorState2.transactionFilter.from(
  modeField,
  (mode) => mode.type === 1 /* Insert */ ? letThrough : (tr) => {
    const userEvent = tr.annotation(Transaction.userEvent);
    if (userEvent == null) {
      return tr;
    }
    if (!userEvent.startsWith("input")) {
      return tr;
    }
    if (userEvent === "input.type.compose.start") {
      return [];
    }
    if (mode.minor !== 2 /* Normal */) {
      return {
        effects: mode.type === 0 /* Normal */ ? MODE_EFF.NORMAL : MODE_EFF.SELECT
      };
    }
    return [];
  }
);
var inputHandler = EditorView.inputHandler.from(
  modeField,
  (mode) => (view, _from, _to, text) => {
    if (mode.type === 1 /* Insert */) {
      return false;
    }
    if (mode.expecting) {
      mode.expecting.callback(view, text, mode.expecting.metadata);
      return true;
    }
    return false;
  }
);
var updateListener = EditorView.updateListener.of((viewUpdate) => {
  const { state, startState } = viewUpdate;
  const panel = getPanel(viewUpdate.view, statusPanel);
  const mode = state.field(modeField);
  const startMode = startState.field(modeField);
  if (mode !== startMode) {
    const startExternalMode = toExternalMode(startMode);
    const externalMode = toExternalMode(mode);
    if (startExternalMode !== externalMode) {
      panel.setMode(externalMode);
    }
  }
  panel.setLineCol();
});
var helixKeymap = keymap.of(toCodemirrorKeymap(helixCommandBindings));
function helix() {
  return [
    EditorView.theme({
      ".cm-cursor": {
        display: "none !important"
      },
      ".cm-hx-cursor": {
        background: "#ccc"
      },
      // WARNING: flaky
      ".cm-searchMatch": {
        background: "initial"
      }
    }),
    drawSelection({
      cursorBlinkRate: 0
    }),
    helixKeymap,
    modeField,
    historyField,
    registerField,
    searchRegisterField,
    changeFilter,
    inputHandler,
    cursorField,
    updateListener,
    showPanel.of(statusPanel),
    showPanel.of(commandPanel),
    search()
  ];
}
function commandPanel(view) {
  return new CommandPanel(view);
}
function statusPanel(view) {
  const dom = el("div");
  dom.style.display = "flex";
  dom.style.justifyContent = "space-between";
  dom.style.fontFamily = "monospace";
  const mode = el("span");
  mode.textContent = "NOR";
  dom.insertBefore(mode, null);
  const pos = el("span");
  dom.insertBefore(pos, null);
  function setLineCol() {
    const { line, column } = lineCol(view);
    pos.textContent = `${line}:${column}`;
  }
  setLineCol();
  return {
    dom,
    setMode(modeStr) {
      mode.textContent = modeStr;
    },
    setLineCol
  };
}
function lineCol(view) {
  const head = view.state.selection.main.head;
  const lineDesc = view.state.doc.lineAt(head);
  const line = lineDesc.number;
  const column = head - lineDesc.from + 1;
  return { line, column };
}
var CommandPanel = class {
  constructor(view) {
    this.view = view;
    this.dom = el("div");
    this.dom.style.display = "flex";
    this.dom.style.justifyContent = "space-between";
    this.dom.style.fontFamily = "monospace";
    this.dom.style.minHeight = "2em";
    this.command = el("span");
    this.input = el("span");
    this.input.style.visibility = "hidden";
    this.showCommand(null);
    {
      const label = el("span");
      label.textContent = "search:";
      this.input.insertBefore(label, null);
    }
    this.dom.insertBefore(this.input, null);
    this.dom.insertBefore(this.command, null);
  }
  dom;
  command;
  input;
  showSearchInput() {
    const input = this.searchInput();
    this.input.insertBefore(input, null);
    this.input.style.visibility = "";
    input.focus();
  }
  showCommand(command) {
    if (command) {
      this.command.textContent = command;
    } else {
      this.command.innerHTML = "&nbsp;";
    }
  }
  searchInput() {
    const { view } = this;
    const input = el("input");
    input.type = "text";
    input.style.border = "none";
    input.style.outline = "none";
    input.style.background = "inherit";
    let isCompositing = false;
    input.addEventListener("compositionstart", () => {
      isCompositing = true;
    });
    input.addEventListener("compositionend", () => {
      isCompositing = false;
    });
    input.addEventListener("blur", () => {
      this.closeSearchPanel(false);
    });
    input.addEventListener("input", () => {
      const query = new SearchQuery({
        search: input.value,
        regexp: true,
        caseSensitive: false
      });
      const effect = setSearchQuery.of(query);
      view.dispatch({ effects: effect });
      addSearch(view, query);
    });
    input.addEventListener("keydown", (event) => {
      if (isCompositing) {
        return;
      }
      const isEnter = event.key === "Enter";
      if (isEnter || event.key === "Escape") {
        this.closeSearchPanel(isEnter);
      }
    });
    return input;
  }
  closeSearchPanel(accept) {
    this.view.dispatch({
      effects: [
        searchEffect.of({
          type: 1 /* Exit */,
          query: accept ? getSearchQuery(this.view.state) : void 0
        }),
        setSearchQuery.of(new SearchQuery({ search: "" }))
      ]
    });
    this.input.removeChild(this.input.lastChild);
    this.input.style.visibility = "hidden";
    this.view.focus();
  }
};
function toExternalMode(mode) {
  switch (mode.type) {
    case 0 /* Normal */:
      return "NOR";
    case 4 /* Select */:
      return "SEL";
    case 1 /* Insert */:
      return "INS";
  }
}
function findText(view, text, {
  inclusive,
  reverse
}) {
  const mode = view.state.field(modeField);
  const select = mode.type === 4 /* Select */;
  const selection = view.state.selection.main;
  const start = reverse ? selection.from : selection.to;
  const doc = view.state.doc.toString();
  const rawIndex = reverse ? doc.lastIndexOf(text, start) : doc.indexOf(text, start);
  const resetEffect = select ? MODE_EFF.SELECT : MODE_EFF.NORMAL;
  if (rawIndex === -1) {
    view.dispatch({
      effects: resetEffect
    });
    return;
  }
  const index = inclusive ? rawIndex : reverse ? rawIndex + 1 : rawIndex - 1;
  const newSelection = EditorSelection2.range(selection.head, index);
  view.dispatch({
    effects: resetEffect,
    selection: newSelection
  });
}
var PAIRS = {
  "(": ["(", ")", true],
  ")": ["(", ")", false],
  "{": ["{", "}", true],
  "}": ["{", "}", false],
  "[": ["[", "]", true],
  "]": ["[", "]", false],
  "<": ["<", ">", true],
  ">": ["<", ">", false]
};
var MATCHEABLE = /* @__PURE__ */ new Set([...Object.keys(PAIRS), `"`, "'"]);
function matchBracket(view) {
  const selection = view.state.selection.main;
  const char = view.state.doc.slice(selection.head, selection.head + 1).toString();
  if (!MATCHEABLE.has(char)) {
    return null;
  }
  const open = PAIRS[char]?.[2] ?? false;
  const match = matchBrackets(
    view.state,
    selection.head + (open ? 0 : 1),
    open ? 1 : -1
  );
  if (match) {
    return match.end;
  }
}
function surround(view, char, proxy) {
  const pair = PAIRS[char];
  const selection = view.state.selection.main;
  const range = helixSelection(selection, view.state.doc);
  const open = pair?.[0] ?? char;
  const close = pair?.[1] ?? char;
  view.dispatch({
    changes: [
      {
        from: range.from,
        insert: open
      },
      {
        from: range.to,
        insert: close
      }
    ],
    effects: MODE_EFF.NORMAL
  });
  const newSelection = proxy.original.state.selection.main;
  const offset = newSelection.anchor === newSelection.from ? 1 : -1;
  const anchor = newSelection.anchor - offset;
  const head = newSelection.head + offset;
  proxy.original.dispatch({
    selection: EditorSelection2.range(anchor, head)
  });
}
function el(tag) {
  return document.createElement(tag);
}
export {
  helix
};
